# coding: utf-8

from collections import OrderedDict

from django.conf import settings
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param


class ThuxPageNumberPagination(PageNumberPagination):
    page_size_query_param = settings.REST_FRAMEWORK.get('PAGE_SIZE_QUERY_PARAM', 'per_page')

    def get_last_link(self):
        url = self.request.build_absolute_uri()
        page_number = self.page.paginator.num_pages
        return replace_query_param(url, self.page_query_param, page_number)

    def get_first_link(self):
        url = self.request.build_absolute_uri()
        return replace_query_param(url, self.page_query_param, 1)

    def get_paginated_response(self, data):
        next = previous = None
        if self.page.has_next():
            next = self.page.number + 1
        if self.page.has_previous():
            previous = self.page.number - 1
        return Response(OrderedDict([
            ('count', self.page.paginator.count),
            ('first', 1),
            ('first_link', self.get_first_link()),
            ('next', next),
            ('next_link', self.get_next_link()),
            ('current', self.page.number),
            ('previous', previous),
            ('previous_link', self.get_previous_link()),
            ('last', self.page.paginator.num_pages),
            ('last_link', self.get_last_link()),
            ('per_page', self.page.paginator.per_page),
            ('results', data)
        ]))

    def get_paginated_response_schema(self, schema):
        base_url = self.get_base_url()
        return {
            'type': 'object',
            'properties': {
                'count': {
                    'type': 'integer',
                    'example': 123,
                },
                'first': {
                    'type': 'integer',
                    'example': 1,
                },
                'first_link': {
                    'type': 'string',
                    'nullable': True,
                    'format': 'uri',
                    'example': '{base_url}/accounts/?{page_query_param}=1'.format(
                        base_url=base_url, page_query_param=self.page_query_param
                    )
                },
                'next': {
                    'type': 'integer',
                    'nullable': True,
                    'example': 2,
                },
                'next_link': {
                    'type': 'string',
                    'nullable': True,
                    'format': 'uri',
                    'example': '{base_url}/accounts/?{page_query_param}=1'.format(
                        base_url=base_url, page_query_param=self.page_query_param
                    )
                },
                'current': {
                    'type': 'integer',
                    'example': 1,
                },
                'previous': {
                    'type': 'integer',
                    'nullable': True,
                    'example': 1,
                },
                'previous_link': {
                    'type': 'string',
                    'nullable': True,
                    'format': 'uri',
                    'example': '{base_url}/accounts/?{page_query_param}=1'.format(
                        base_url=base_url, page_query_param=self.page_query_param
                    )
                },
                'last': {
                    'type': 'integer',
                    'nullable': True,
                    'example': 3,
                },
                'last_link': {
                    'type': 'string',
                    'nullable': True,
                    'format': 'uri',
                    'example': '{base_url}/accounts/?{page_query_param}=3'.format(
                        base_url=base_url, page_query_param=self.page_query_param,
                    )
                },
                'per_page': {
                    'type': 'integer',
                    'example': 10,
                },
                'results': schema,
            },
        }

    @staticmethod
    def get_base_url():
        base_protocol = getattr(settings, 'BASE_PROTOCOL', 'http://')
        base_url = getattr(settings, 'BASE_URL', 'example.thux.dev')
        return f'{base_protocol}{base_url}'
