from collections import defaultdict

from django.conf.urls import include, url
from django.urls import path
from drf_yasg import openapi
from drf_yasg.generators import OpenAPISchemaGenerator
from drf_yasg.inspectors import SwaggerAutoSchema
from drf_yasg.renderers import OpenAPIRenderer, ReDocRenderer
from drf_yasg.views import get_schema_view
from rest_framework import permissions
from web.settings import DRF_YASG_SCHEMA_VIEW


class ThuxSwaggerAutoSchema(SwaggerAutoSchema):
    def get_summary_and_description(self):
        summary, description = super().get_summary_and_description()
        if not summary:
            summary = self.view.get_view_name()
        return summary, description


class ThuxReDocRenderer(ReDocRenderer):
    template = 'drf-yasg/thux_redoc.html'
    format = 'redoc'


thux_renderer_classes = (OpenAPIRenderer, ThuxReDocRenderer)


class SuperuserOpenAPISchemaGenerator(OpenAPISchemaGenerator):
    """
    Override the OpenAPISchemaGenerator to filter out PUT methods (DRF always uses PATCH)
    and to filter endpoints, including only the ones for "editor" profile
    """

    def get_endpoints(self, request):
        enumerator = self.endpoint_enumerator_class(self._gen.patterns, self._gen.urlconf, request=request)
        endpoints = enumerator.get_api_endpoints()
        view_paths = defaultdict(list)
        view_cls = {}
        for path, method, callback in endpoints:
            if method == 'PUT':
                continue
            path = path.replace('{profile}', 'superuser')
            view = self.create_view(callback, method, request)
            view.kwargs = {'profile': 'superuser'}
            path = self.coerce_path(path, view)
            view_paths[path].append((method, view))
            view_cls[path] = callback.cls
        return {path: (view_cls[path], methods) for path, methods in view_paths.items()}


superuser_schema_view = get_schema_view(
    openapi.Info(
        title=DRF_YASG_SCHEMA_VIEW['title'],
        default_version=DRF_YASG_SCHEMA_VIEW['default_version'],
        description=DRF_YASG_SCHEMA_VIEW['description'],
    ),
    patterns=[
        path('api/rest-auth/', include('rest_auth.urls')),
        url(r'^api/v1/account/', include('apps.account.api.version1.urls')),
        #url(r'^api/address/', include('apps.address.api.urls')),
        #url(r'^api/v1/agenda/', include('apps.agenda.api.version1.urls')),
        #url(r'^api/v1/bank/', include('apps.bank.api.version1.urls')),
        # url(r'^api/v1/carriage/', include('apps.carriage.api.version1.urls')),
        # url(r'^api/v1/cron/', include('apps.cron.api.version1.urls')),
        # url(r'^api/v1/currency/', include('apps.currency.api.version1.urls')),
        # url(r'^api/v1/i18n/', include('apps.i18n.api.version1.urls')),
        # url(r'^api/v1/iso/', include('apps.iso.api.version1.urls')),
        # url(r'^api/v1/l10n/', include('apps.l10n.api.version1.urls')),
        # url(r'^api/v1/miscellaneous/', include('apps.miscellaneous.api.version1.urls')),
        # url(r'^api/v1/news/', include('apps.news.api.version1.urls')),
        # url(r'^api/v1/notify/', include('apps.notify.api.version1.urls')),
        # url(r'^api/v1/office/', include('apps.office.api.version1.urls')),
        # url(r'^api/v1/order/', include('apps.order.api.version1.urls')),
        # url(r'^api/v1/organization/', include('apps.organization.api.version1.urls')),
        # url(r'^api/v1/payment/', include('apps.payment.api.version1.urls')),
        # url(r'^api/v1/price/', include('apps.price.api.version1.urls')),
        # url(r'^api/v1/product/', include('apps.product.api.version1.urls')),
        # url(r'^api/v1/project/', include('apps.project.api.version1.urls')),
        # url(r'^api/v1/quotation/', include('apps.quotation.api.version1.urls')),
        # url(r'^api/v1/store/', include('apps.store.api.version1.urls')),
        url(r'^api/v1/ticket/', include('apps.ticket.api.version1.urls')),
        url(r'^api/v1/timereport/', include('apps.timereport.api.version1.urls')),
        #url(r'^api/v1/vat/', include('apps.vat.api.version1.urls')),

    ],
    public=True,
    permission_classes=(permissions.AllowAny,),
    generator_class=SuperuserOpenAPISchemaGenerator,
)


class ManagerOpenAPISchemaGenerator(OpenAPISchemaGenerator):
    """
    Override the OpenAPISchemaGenerator to filter out PUT methods (DRF always uses PATCH)
    and to filter endpoints, including only the ones for "editor" profile
    """

    def get_endpoints(self, request):
        enumerator = self.endpoint_enumerator_class(self._gen.patterns, self._gen.urlconf, request=request)
        endpoints = enumerator.get_api_endpoints()
        view_paths = defaultdict(list)
        view_cls = {}
        for path, method, callback in endpoints:
            if method == 'PUT':
                continue
            path = path.replace('{profile}', 'manager')
            view = self.create_view(callback, method, request)
            view.kwargs = {'profile': 'manager'}
            path = self.coerce_path(path, view)
            view_paths[path].append((method, view))
            view_cls[path] = callback.cls
        return {path: (view_cls[path], methods) for path, methods in view_paths.items()}


manager_schema_view = get_schema_view(
    openapi.Info(
        title=DRF_YASG_SCHEMA_VIEW['title'],
        default_version=DRF_YASG_SCHEMA_VIEW['default_version'],
        description=DRF_YASG_SCHEMA_VIEW['description'],
    ),
    patterns=[
        path('api/rest-auth/', include('rest_auth.urls')),
        url(r'^api/v1/account/', include('apps.account.api.version1.urls')),
        url(r'^api/v1/ticket/', include('apps.ticket.api.version1.urls')),
        url(r'^api/v1/timereport/', include('apps.timereport.api.version1.urls')),
    ],
    public=True,
    permission_classes=(permissions.AllowAny,),
    generator_class=ManagerOpenAPISchemaGenerator,
)


class WorkerOpenAPISchemaGenerator(OpenAPISchemaGenerator):
    """
    Override the OpenAPISchemaGenerator to filter out PUT methods (DRF always uses PATCH)
    and to filter endpoints, including only the ones for "editor" profile
    """

    def get_endpoints(self, request):
        enumerator = self.endpoint_enumerator_class(self._gen.patterns, self._gen.urlconf, request=request)
        endpoints = enumerator.get_api_endpoints()
        view_paths = defaultdict(list)
        view_cls = {}
        for path, method, callback in endpoints:
            if method == 'PUT':
                continue
            path = path.replace('{profile}', 'worker')
            view = self.create_view(callback, method, request)
            view.kwargs = {'profile': 'worker'}
            path = self.coerce_path(path, view)
            view_paths[path].append((method, view))
            view_cls[path] = callback.cls
        return {path: (view_cls[path], methods) for path, methods in view_paths.items()}


worker_schema_view = get_schema_view(
    openapi.Info(
        title=DRF_YASG_SCHEMA_VIEW['title'],
        default_version=DRF_YASG_SCHEMA_VIEW['default_version'],
        description=DRF_YASG_SCHEMA_VIEW['description'],
    ),
    patterns=[
        path('api/rest-auth/', include('rest_auth.urls')),
        url(r'^api/v1/account/', include('apps.account.api.version1.urls')),
        url(r'^api/v1/ticket/', include('apps.ticket.api.version1.urls')),
        url(r'^api/v1/timereport/', include('apps.timereport.api.version1.urls')),
    ],
    public=True,
    permission_classes=(permissions.AllowAny,),
    generator_class=ManagerOpenAPISchemaGenerator,
)

